#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 2
#endif
#ifndef PFXDIST
   #define PFXDIST 0
#endif
#ifndef PFADIST
   #define PFADIST 0
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGEMV(ATL_CINT M, ATL_CINT N, const TYPE *A, ATL_CINT lda,
               const TYPE *X, TYPE *Y)
/*
 * 112x2 unrolled mvt_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N2=(N/2)*2, M112=(M/112)*112, lda2=lda*2;
   ATL_INT j;

   for (j=N2; j; j -= 2, A += lda2, Y += 2)
   {
      const double *A0=A, *A1=A0+lda;
      register double y0=ATL_rzero, y1=ATL_rzero;
      ATL_INT i;
      prefY(Y+2+2-1);
      for (i=0; i < M112; i += 112)
      {
         const register double x0=X[i+0], x1=X[i+1], x2=X[i+2], x3=X[i+3],
                               x4=X[i+4], x5=X[i+5], x6=X[i+6], x7=X[i+7],
                               x8=X[i+8], x9=X[i+9], x10=X[i+10], x11=X[i+11],
                               x12=X[i+12], x13=X[i+13], x14=X[i+14],
                               x15=X[i+15], x16=X[i+16], x17=X[i+17],
                               x18=X[i+18], x19=X[i+19], x20=X[i+20],
                               x21=X[i+21], x22=X[i+22], x23=X[i+23],
                               x24=X[i+24], x25=X[i+25], x26=X[i+26],
                               x27=X[i+27], x28=X[i+28], x29=X[i+29],
                               x30=X[i+30], x31=X[i+31], x32=X[i+32],
                               x33=X[i+33], x34=X[i+34], x35=X[i+35],
                               x36=X[i+36], x37=X[i+37], x38=X[i+38],
                               x39=X[i+39], x40=X[i+40], x41=X[i+41],
                               x42=X[i+42], x43=X[i+43], x44=X[i+44],
                               x45=X[i+45], x46=X[i+46], x47=X[i+47],
                               x48=X[i+48], x49=X[i+49], x50=X[i+50],
                               x51=X[i+51], x52=X[i+52], x53=X[i+53],
                               x54=X[i+54], x55=X[i+55], x56=X[i+56],
                               x57=X[i+57], x58=X[i+58], x59=X[i+59],
                               x60=X[i+60], x61=X[i+61], x62=X[i+62],
                               x63=X[i+63], x64=X[i+64], x65=X[i+65],
                               x66=X[i+66], x67=X[i+67], x68=X[i+68],
                               x69=X[i+69], x70=X[i+70], x71=X[i+71],
                               x72=X[i+72], x73=X[i+73], x74=X[i+74],
                               x75=X[i+75], x76=X[i+76], x77=X[i+77],
                               x78=X[i+78], x79=X[i+79], x80=X[i+80],
                               x81=X[i+81], x82=X[i+82], x83=X[i+83],
                               x84=X[i+84], x85=X[i+85], x86=X[i+86],
                               x87=X[i+87], x88=X[i+88], x89=X[i+89],
                               x90=X[i+90], x91=X[i+91], x92=X[i+92],
                               x93=X[i+93], x94=X[i+94], x95=X[i+95],
                               x96=X[i+96], x97=X[i+97], x98=X[i+98],
                               x99=X[i+99], x100=X[i+100], x101=X[i+101],
                               x102=X[i+102], x103=X[i+103], x104=X[i+104],
                               x105=X[i+105], x106=X[i+106], x107=X[i+107],
                               x108=X[i+108], x109=X[i+109], x110=X[i+110],
                               x111=X[i+111];

         prefX(X);
         y0 += A0[i+0] * x0;
         prefA(A0);
         y1 += A1[i+0] * x0;
         prefA(A1);
         y0 += A0[i+1] * x1;
         y1 += A1[i+1] * x1;
         y0 += A0[i+2] * x2;
         y1 += A1[i+2] * x2;
         y0 += A0[i+3] * x3;
         y1 += A1[i+3] * x3;
         y0 += A0[i+4] * x4;
         y1 += A1[i+4] * x4;
         y0 += A0[i+5] * x5;
         y1 += A1[i+5] * x5;
         y0 += A0[i+6] * x6;
         y1 += A1[i+6] * x6;
         y0 += A0[i+7] * x7;
         y1 += A1[i+7] * x7;
         y0 += A0[i+8] * x8;
         prefA(A0);
         y1 += A1[i+8] * x8;
         prefA(A1);
         y0 += A0[i+9] * x9;
         y1 += A1[i+9] * x9;
         y0 += A0[i+10] * x10;
         y1 += A1[i+10] * x10;
         y0 += A0[i+11] * x11;
         y1 += A1[i+11] * x11;
         y0 += A0[i+12] * x12;
         y1 += A1[i+12] * x12;
         y0 += A0[i+13] * x13;
         y1 += A1[i+13] * x13;
         y0 += A0[i+14] * x14;
         y1 += A1[i+14] * x14;
         y0 += A0[i+15] * x15;
         y1 += A1[i+15] * x15;
         y0 += A0[i+16] * x16;
         prefA(A0);
         y1 += A1[i+16] * x16;
         prefA(A1);
         y0 += A0[i+17] * x17;
         y1 += A1[i+17] * x17;
         y0 += A0[i+18] * x18;
         y1 += A1[i+18] * x18;
         y0 += A0[i+19] * x19;
         y1 += A1[i+19] * x19;
         y0 += A0[i+20] * x20;
         y1 += A1[i+20] * x20;
         y0 += A0[i+21] * x21;
         y1 += A1[i+21] * x21;
         y0 += A0[i+22] * x22;
         y1 += A1[i+22] * x22;
         y0 += A0[i+23] * x23;
         y1 += A1[i+23] * x23;
         y0 += A0[i+24] * x24;
         prefA(A0);
         y1 += A1[i+24] * x24;
         prefA(A1);
         y0 += A0[i+25] * x25;
         y1 += A1[i+25] * x25;
         y0 += A0[i+26] * x26;
         y1 += A1[i+26] * x26;
         y0 += A0[i+27] * x27;
         y1 += A1[i+27] * x27;
         y0 += A0[i+28] * x28;
         y1 += A1[i+28] * x28;
         y0 += A0[i+29] * x29;
         y1 += A1[i+29] * x29;
         y0 += A0[i+30] * x30;
         y1 += A1[i+30] * x30;
         y0 += A0[i+31] * x31;
         y1 += A1[i+31] * x31;
         y0 += A0[i+32] * x32;
         prefA(A0);
         y1 += A1[i+32] * x32;
         prefA(A1);
         y0 += A0[i+33] * x33;
         y1 += A1[i+33] * x33;
         y0 += A0[i+34] * x34;
         y1 += A1[i+34] * x34;
         y0 += A0[i+35] * x35;
         y1 += A1[i+35] * x35;
         y0 += A0[i+36] * x36;
         y1 += A1[i+36] * x36;
         y0 += A0[i+37] * x37;
         y1 += A1[i+37] * x37;
         y0 += A0[i+38] * x38;
         y1 += A1[i+38] * x38;
         y0 += A0[i+39] * x39;
         y1 += A1[i+39] * x39;
         y0 += A0[i+40] * x40;
         prefA(A0);
         y1 += A1[i+40] * x40;
         prefA(A1);
         y0 += A0[i+41] * x41;
         y1 += A1[i+41] * x41;
         y0 += A0[i+42] * x42;
         y1 += A1[i+42] * x42;
         y0 += A0[i+43] * x43;
         y1 += A1[i+43] * x43;
         y0 += A0[i+44] * x44;
         y1 += A1[i+44] * x44;
         y0 += A0[i+45] * x45;
         y1 += A1[i+45] * x45;
         y0 += A0[i+46] * x46;
         y1 += A1[i+46] * x46;
         y0 += A0[i+47] * x47;
         y1 += A1[i+47] * x47;
         y0 += A0[i+48] * x48;
         prefA(A0);
         y1 += A1[i+48] * x48;
         prefA(A1);
         y0 += A0[i+49] * x49;
         y1 += A1[i+49] * x49;
         y0 += A0[i+50] * x50;
         y1 += A1[i+50] * x50;
         y0 += A0[i+51] * x51;
         y1 += A1[i+51] * x51;
         y0 += A0[i+52] * x52;
         y1 += A1[i+52] * x52;
         y0 += A0[i+53] * x53;
         y1 += A1[i+53] * x53;
         y0 += A0[i+54] * x54;
         y1 += A1[i+54] * x54;
         y0 += A0[i+55] * x55;
         y1 += A1[i+55] * x55;
         y0 += A0[i+56] * x56;
         prefA(A0);
         y1 += A1[i+56] * x56;
         prefA(A1);
         y0 += A0[i+57] * x57;
         y1 += A1[i+57] * x57;
         y0 += A0[i+58] * x58;
         y1 += A1[i+58] * x58;
         y0 += A0[i+59] * x59;
         y1 += A1[i+59] * x59;
         y0 += A0[i+60] * x60;
         y1 += A1[i+60] * x60;
         y0 += A0[i+61] * x61;
         y1 += A1[i+61] * x61;
         y0 += A0[i+62] * x62;
         y1 += A1[i+62] * x62;
         y0 += A0[i+63] * x63;
         y1 += A1[i+63] * x63;
         y0 += A0[i+64] * x64;
         prefA(A0);
         y1 += A1[i+64] * x64;
         prefA(A1);
         y0 += A0[i+65] * x65;
         y1 += A1[i+65] * x65;
         y0 += A0[i+66] * x66;
         y1 += A1[i+66] * x66;
         y0 += A0[i+67] * x67;
         y1 += A1[i+67] * x67;
         y0 += A0[i+68] * x68;
         y1 += A1[i+68] * x68;
         y0 += A0[i+69] * x69;
         y1 += A1[i+69] * x69;
         y0 += A0[i+70] * x70;
         y1 += A1[i+70] * x70;
         y0 += A0[i+71] * x71;
         y1 += A1[i+71] * x71;
         y0 += A0[i+72] * x72;
         prefA(A0);
         y1 += A1[i+72] * x72;
         prefA(A1);
         y0 += A0[i+73] * x73;
         y1 += A1[i+73] * x73;
         y0 += A0[i+74] * x74;
         y1 += A1[i+74] * x74;
         y0 += A0[i+75] * x75;
         y1 += A1[i+75] * x75;
         y0 += A0[i+76] * x76;
         y1 += A1[i+76] * x76;
         y0 += A0[i+77] * x77;
         y1 += A1[i+77] * x77;
         y0 += A0[i+78] * x78;
         y1 += A1[i+78] * x78;
         y0 += A0[i+79] * x79;
         y1 += A1[i+79] * x79;
         y0 += A0[i+80] * x80;
         prefA(A0);
         y1 += A1[i+80] * x80;
         prefA(A1);
         y0 += A0[i+81] * x81;
         y1 += A1[i+81] * x81;
         y0 += A0[i+82] * x82;
         y1 += A1[i+82] * x82;
         y0 += A0[i+83] * x83;
         y1 += A1[i+83] * x83;
         y0 += A0[i+84] * x84;
         y1 += A1[i+84] * x84;
         y0 += A0[i+85] * x85;
         y1 += A1[i+85] * x85;
         y0 += A0[i+86] * x86;
         y1 += A1[i+86] * x86;
         y0 += A0[i+87] * x87;
         y1 += A1[i+87] * x87;
         y0 += A0[i+88] * x88;
         prefA(A0);
         y1 += A1[i+88] * x88;
         prefA(A1);
         y0 += A0[i+89] * x89;
         y1 += A1[i+89] * x89;
         y0 += A0[i+90] * x90;
         y1 += A1[i+90] * x90;
         y0 += A0[i+91] * x91;
         y1 += A1[i+91] * x91;
         y0 += A0[i+92] * x92;
         y1 += A1[i+92] * x92;
         y0 += A0[i+93] * x93;
         y1 += A1[i+93] * x93;
         y0 += A0[i+94] * x94;
         y1 += A1[i+94] * x94;
         y0 += A0[i+95] * x95;
         y1 += A1[i+95] * x95;
         y0 += A0[i+96] * x96;
         prefA(A0);
         y1 += A1[i+96] * x96;
         prefA(A1);
         y0 += A0[i+97] * x97;
         y1 += A1[i+97] * x97;
         y0 += A0[i+98] * x98;
         y1 += A1[i+98] * x98;
         y0 += A0[i+99] * x99;
         y1 += A1[i+99] * x99;
         y0 += A0[i+100] * x100;
         y1 += A1[i+100] * x100;
         y0 += A0[i+101] * x101;
         y1 += A1[i+101] * x101;
         y0 += A0[i+102] * x102;
         y1 += A1[i+102] * x102;
         y0 += A0[i+103] * x103;
         y1 += A1[i+103] * x103;
         y0 += A0[i+104] * x104;
         prefA(A0);
         y1 += A1[i+104] * x104;
         prefA(A1);
         y0 += A0[i+105] * x105;
         y1 += A1[i+105] * x105;
         y0 += A0[i+106] * x106;
         y1 += A1[i+106] * x106;
         y0 += A0[i+107] * x107;
         y1 += A1[i+107] * x107;
         y0 += A0[i+108] * x108;
         y1 += A1[i+108] * x108;
         y0 += A0[i+109] * x109;
         y1 += A1[i+109] * x109;
         y0 += A0[i+110] * x110;
         y1 += A1[i+110] * x110;
         y0 += A0[i+111] * x111;
         y1 += A1[i+111] * x111;
      }
      for (i=M112; i < M; i++)
      {
         const register double x0 = X[i];
         y0 += A0[i] * x0;
         y1 += A1[i] * x0;
      }
      #ifdef BETA0
         Y[0] = y0;
         Y[1] = y1;
      #else
         Y[0] += y0;
         Y[1] += y1;
      #endif
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N2; j; j--, A += lda, Y++)
   {
      const double *A0=A;
      register double y0=ATL_rzero;
      ATL_INT i;
      prefY(Y+1+1-1);
      for (i=0; i < M112; i += 112)
      {
         const register double x0=X[i+0], x1=X[i+1], x2=X[i+2], x3=X[i+3],
                               x4=X[i+4], x5=X[i+5], x6=X[i+6], x7=X[i+7],
                               x8=X[i+8], x9=X[i+9], x10=X[i+10], x11=X[i+11],
                               x12=X[i+12], x13=X[i+13], x14=X[i+14],
                               x15=X[i+15], x16=X[i+16], x17=X[i+17],
                               x18=X[i+18], x19=X[i+19], x20=X[i+20],
                               x21=X[i+21], x22=X[i+22], x23=X[i+23],
                               x24=X[i+24], x25=X[i+25], x26=X[i+26],
                               x27=X[i+27], x28=X[i+28], x29=X[i+29],
                               x30=X[i+30], x31=X[i+31], x32=X[i+32],
                               x33=X[i+33], x34=X[i+34], x35=X[i+35],
                               x36=X[i+36], x37=X[i+37], x38=X[i+38],
                               x39=X[i+39], x40=X[i+40], x41=X[i+41],
                               x42=X[i+42], x43=X[i+43], x44=X[i+44],
                               x45=X[i+45], x46=X[i+46], x47=X[i+47],
                               x48=X[i+48], x49=X[i+49], x50=X[i+50],
                               x51=X[i+51], x52=X[i+52], x53=X[i+53],
                               x54=X[i+54], x55=X[i+55], x56=X[i+56],
                               x57=X[i+57], x58=X[i+58], x59=X[i+59],
                               x60=X[i+60], x61=X[i+61], x62=X[i+62],
                               x63=X[i+63], x64=X[i+64], x65=X[i+65],
                               x66=X[i+66], x67=X[i+67], x68=X[i+68],
                               x69=X[i+69], x70=X[i+70], x71=X[i+71],
                               x72=X[i+72], x73=X[i+73], x74=X[i+74],
                               x75=X[i+75], x76=X[i+76], x77=X[i+77],
                               x78=X[i+78], x79=X[i+79], x80=X[i+80],
                               x81=X[i+81], x82=X[i+82], x83=X[i+83],
                               x84=X[i+84], x85=X[i+85], x86=X[i+86],
                               x87=X[i+87], x88=X[i+88], x89=X[i+89],
                               x90=X[i+90], x91=X[i+91], x92=X[i+92],
                               x93=X[i+93], x94=X[i+94], x95=X[i+95],
                               x96=X[i+96], x97=X[i+97], x98=X[i+98],
                               x99=X[i+99], x100=X[i+100], x101=X[i+101],
                               x102=X[i+102], x103=X[i+103], x104=X[i+104],
                               x105=X[i+105], x106=X[i+106], x107=X[i+107],
                               x108=X[i+108], x109=X[i+109], x110=X[i+110],
                               x111=X[i+111];

         prefX(X);
         y0 += A0[i+0] * x0;
         prefA(A0);
         y0 += A0[i+1] * x1;
         y0 += A0[i+2] * x2;
         y0 += A0[i+3] * x3;
         y0 += A0[i+4] * x4;
         y0 += A0[i+5] * x5;
         y0 += A0[i+6] * x6;
         y0 += A0[i+7] * x7;
         y0 += A0[i+8] * x8;
         prefA(A0);
         y0 += A0[i+9] * x9;
         y0 += A0[i+10] * x10;
         y0 += A0[i+11] * x11;
         y0 += A0[i+12] * x12;
         y0 += A0[i+13] * x13;
         y0 += A0[i+14] * x14;
         y0 += A0[i+15] * x15;
         y0 += A0[i+16] * x16;
         prefA(A0);
         y0 += A0[i+17] * x17;
         y0 += A0[i+18] * x18;
         y0 += A0[i+19] * x19;
         y0 += A0[i+20] * x20;
         y0 += A0[i+21] * x21;
         y0 += A0[i+22] * x22;
         y0 += A0[i+23] * x23;
         y0 += A0[i+24] * x24;
         prefA(A0);
         y0 += A0[i+25] * x25;
         y0 += A0[i+26] * x26;
         y0 += A0[i+27] * x27;
         y0 += A0[i+28] * x28;
         y0 += A0[i+29] * x29;
         y0 += A0[i+30] * x30;
         y0 += A0[i+31] * x31;
         y0 += A0[i+32] * x32;
         prefA(A0);
         y0 += A0[i+33] * x33;
         y0 += A0[i+34] * x34;
         y0 += A0[i+35] * x35;
         y0 += A0[i+36] * x36;
         y0 += A0[i+37] * x37;
         y0 += A0[i+38] * x38;
         y0 += A0[i+39] * x39;
         y0 += A0[i+40] * x40;
         prefA(A0);
         y0 += A0[i+41] * x41;
         y0 += A0[i+42] * x42;
         y0 += A0[i+43] * x43;
         y0 += A0[i+44] * x44;
         y0 += A0[i+45] * x45;
         y0 += A0[i+46] * x46;
         y0 += A0[i+47] * x47;
         y0 += A0[i+48] * x48;
         prefA(A0);
         y0 += A0[i+49] * x49;
         y0 += A0[i+50] * x50;
         y0 += A0[i+51] * x51;
         y0 += A0[i+52] * x52;
         y0 += A0[i+53] * x53;
         y0 += A0[i+54] * x54;
         y0 += A0[i+55] * x55;
         y0 += A0[i+56] * x56;
         prefA(A0);
         y0 += A0[i+57] * x57;
         y0 += A0[i+58] * x58;
         y0 += A0[i+59] * x59;
         y0 += A0[i+60] * x60;
         y0 += A0[i+61] * x61;
         y0 += A0[i+62] * x62;
         y0 += A0[i+63] * x63;
         y0 += A0[i+64] * x64;
         prefA(A0);
         y0 += A0[i+65] * x65;
         y0 += A0[i+66] * x66;
         y0 += A0[i+67] * x67;
         y0 += A0[i+68] * x68;
         y0 += A0[i+69] * x69;
         y0 += A0[i+70] * x70;
         y0 += A0[i+71] * x71;
         y0 += A0[i+72] * x72;
         prefA(A0);
         y0 += A0[i+73] * x73;
         y0 += A0[i+74] * x74;
         y0 += A0[i+75] * x75;
         y0 += A0[i+76] * x76;
         y0 += A0[i+77] * x77;
         y0 += A0[i+78] * x78;
         y0 += A0[i+79] * x79;
         y0 += A0[i+80] * x80;
         prefA(A0);
         y0 += A0[i+81] * x81;
         y0 += A0[i+82] * x82;
         y0 += A0[i+83] * x83;
         y0 += A0[i+84] * x84;
         y0 += A0[i+85] * x85;
         y0 += A0[i+86] * x86;
         y0 += A0[i+87] * x87;
         y0 += A0[i+88] * x88;
         prefA(A0);
         y0 += A0[i+89] * x89;
         y0 += A0[i+90] * x90;
         y0 += A0[i+91] * x91;
         y0 += A0[i+92] * x92;
         y0 += A0[i+93] * x93;
         y0 += A0[i+94] * x94;
         y0 += A0[i+95] * x95;
         y0 += A0[i+96] * x96;
         prefA(A0);
         y0 += A0[i+97] * x97;
         y0 += A0[i+98] * x98;
         y0 += A0[i+99] * x99;
         y0 += A0[i+100] * x100;
         y0 += A0[i+101] * x101;
         y0 += A0[i+102] * x102;
         y0 += A0[i+103] * x103;
         y0 += A0[i+104] * x104;
         prefA(A0);
         y0 += A0[i+105] * x105;
         y0 += A0[i+106] * x106;
         y0 += A0[i+107] * x107;
         y0 += A0[i+108] * x108;
         y0 += A0[i+109] * x109;
         y0 += A0[i+110] * x110;
         y0 += A0[i+111] * x111;
      }
      for (i=M112; i < M; i++)
      {
         const register double x0 = X[i];
         y0 += A0[i] * x0;
      }
      #ifdef BETA0
         Y[0] = y0;
      #else
         Y[0] += y0;
      #endif
   }
}
